#pragma once

#include "Mesh.h"
#include "Shader.h"
#include "Renderable.h"

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/matrix_decompose.hpp>

static constexpr float BallDiameter = 0.05715f*2.f;
static constexpr float BallRadius = BallDiameter/2.f;

/**
 * @brief A renderable body with a mesh visualization.
 */
class Body : public Renderable
{
public:
    /**
     * The Body constructor initializes a vector of shared pointers to Mesh objects using the provided mesh
     * paths and sets a flag to indicate whether dynamics should be simulated.
     * 
     * @param meshPathes A vector of const char pointers representing the paths to the mesh files.
     * @param dynamics The "dynamics" parameter is a boolean value that determines whether or not to
     * simulate dynamics for the body. If it is set to true, dynamics will be simulated. If it is set to
     * false, dynamics will not be simulated.
     */
    Body( std::vector<const char *> meshPathes, bool dynamics = true )
        : simulateDynamics( dynamics )
    {
        for (auto path : meshPathes)
            m_meshes.push_back( std::make_shared<Mesh>(path) );
    };

    /**
     * @brief Destroy the Body object
     */
    virtual ~Body() {};

    virtual void render( Shader &shader ) override;

    /**
     * @brief Set the velocity
     * @param velocity The new velocity of the body in world space.
     */
    void setVelocity( glm::vec3 velocity ) { m_velocity = velocity; };

    /**
     * @brief Set the position, as well as update the transformation matrix.
     * @param translation The new position of the body in world space.
     */
    void setPosition( glm::vec3 translation ) { m_position = translation; m_modelToWorld_translation = glm::translate(translation); updateModelToWorld(); };

    /**
     * @brief Set the rotation, as well as update the transformation matrix.
     * @param angle The angle of the new rotation of the body in degrees.
     * @param axis The axis of the new rotation of the body in world space.
     */
    void setRotation( float angle, glm::vec3 axis ) { m_modelToWorld_rotation = glm::rotate(angle,axis); updateModelToWorld(); };

    /**
     * @brief Set the rotation, as well as update the transformation matrix.
     * @param rotation The new rotation of the body in world space.
     */
    void setRotation( glm::mat4 rotation ) { m_modelToWorld_rotation = rotation; updateModelToWorld(); };

    /**
     * @brief Set the scale, as well as update the transformation matrix.
     * @param scale The new scale of the body in world space.
     */
    void setScale( glm::vec3 scale ) { m_modelToWorld_scale = glm::scale(scale); updateModelToWorld(); };

    /** 
     * @brief Directly set the model to world matrix.
     * @note Also tries to update the transformation parts by decomposing
     *       the combined transformation in parts. Reliability unclear.
    */
    void setModelToWorld( glm::mat4 WfM ) {
        glm::vec3 scale;
        glm::quat rotation;
        glm::vec3 translation;
        glm::vec3 skew;
        glm::vec4 perspective;
        glm::decompose(WfM, scale, rotation, translation, skew, perspective);
        m_modelToWorld = WfM;
        m_modelToWorld_rotation = glm::mat4(glm::conjugate(rotation));
        m_modelToWorld_translation = glm::translate(translation);
        m_position = translation;
        m_modelToWorld_scale = glm::scale(scale);
    };

    /**
     * @brief Get the position of the body's origin in world space.
     * @return The position of the body's origin in world space.
     */
    glm::vec3 getPosition() { return m_position; };

    /**
     * @brief Get the position of the body's origin in world space.
     * @return The position of the body's origin in world space.
     */
    glm::vec3 getVelocity() { return m_velocity; };
    
    /**
     * @brief Get the translational part of the transformation.
     * @return The translational part of the transformation.
     */
    glm::mat4 getTranslation() { return m_modelToWorld_translation; };

    /**
     * @brief Get the rotational part of the transformation.
     * @return The rotational part of the transformation.
     */
    glm::mat4 getRotation() { return m_modelToWorld_rotation; };

    /**
     * @brief Get the scaling part of the transformation.
     * @return The scaling part of the transformation.
     */
    glm::mat4 getScale() { return m_modelToWorld_scale; };

    /**
     * @brief Get the Model To World transformation.
     * @return The model to world transformation.
     * @ref m_modelToWorld for what this transformation does.
     */
    glm::mat4 getModelToWorld() { return m_modelToWorld; };

    /**
     * @brief The `getMeshAt` function is a member function of the `Body`
     * class. It returns a shared pointer to the `Mesh` object at the specified
     * index `i` in the `m_meshes` vector.
     * 
     * @param i The index of the mesh that you want to access.
     * @return std::shared_ptr<Mesh> The mesh object pointer.
     */
    std::shared_ptr<Mesh> getMeshAt(int i) { return m_meshes.at(i); };

    /** The mass of the body. */
    float m_mass = 1.f;

    /** The inverse mass of the body (auxiliary variable). */
    float m_invMass = 1.f;

    /** The ball radius. */
    float m_radius = BallRadius;

    /** A flag that determine whether this body's motion is simulated, or it is static. */
    bool simulateDynamics = true;

private:
    /** The meshes that are rendered at the body's position. */
    std::vector<std::shared_ptr<Mesh>> m_meshes;

    /** The translational part of the body's world from model transformation. */
    glm::mat4 m_modelToWorld_translation = glm::mat4( 1.0 );

    /** The rotational part of the body's world from model transformation. */
    glm::mat4 m_modelToWorld_rotation = glm::mat4( 1.0 );

    /** The scaling part of the body's world from model transformation. */
    glm::mat4 m_modelToWorld_scale = glm::mat4( 1.0 );

    /**
     * @brief The body's world from model transformation.
     * 
     * When multiplied with a point in model space (such as given in an
     * .obj-file) from the left side, the point will be transformed in
     * to world space, based on the current position of the body.
     */
    glm::mat4 m_modelToWorld = glm::mat4( 1.0 );

    /** The position of the body's origin in world space. */
    glm::vec3 m_position = glm::vec3( 0 );

    /** The translational velocity of the body's origin in world space. */
    glm::vec3 m_velocity = glm::vec3( 0 );

    /**
     * @brief Calculates the world from model transformation based the
     *        current member values of the transformation parts.
     * @note This might override a previously directly written modelToWorld
     *       matrix, as the parts of the transformation might not have been
     *       extracted properly.
     */
    void updateModelToWorld() { m_modelToWorld = m_modelToWorld_translation * m_modelToWorld_rotation * m_modelToWorld_scale; };
};