#define IMGUI_DEFINE_MATH_OPERATORS
#include <imgui_internal.h>
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/orthonormalize.hpp>

#include <iostream>
#include <memory>
#include <cmath>
#include <algorithm>

#include "Shader.h"
#include "Mesh.h"
#include "Body.h"
#include "Scene.h"
#include "Physics.h"
#include "Camera.h"

static const std::string TableFeltObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/TableFelt.obj") );
static const std::string TableWoodObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/TableWood.obj") );
// static const std::string SphereObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/sphere.obj") );
static const std::string SphereObjFile = std::string( std::string(CMAKE_SOURCE_DIR)+std::string("/models/sphere_highres.obj") );

/**
 * Initializes the application including the physics & GUI.
 */
int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
    // Enable AA
    glfwWindowHint(GLFW_SAMPLES, 4);
    glfwWindowHint(GLFW_MAXIMIZED, true);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(640, 640, "Computergraphik: Billard", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    // Enable VSync
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    ImGui::StyleColorsClassic();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf", 24.0f);

    // Enable depth test:
    glEnable(GL_DEPTH_TEST);

    // Enable anti-aliasing
    glEnable(GL_MULTISAMPLE);

    // *** Create our scene ***
    auto scene = std::make_shared<Scene>();

    // Placing a table at the origin
    auto table = std::make_shared<Body>( std::vector<const char*>{TableFeltObjFile.c_str(), TableWoodObjFile.c_str()}, false );
    table->setRotation( glm::rotate((float)(M_PI/2.f), glm::normalize(glm::vec3(1,0,0))) * glm::rotate((float)(M_PI/2.f), glm::normalize(glm::vec3(0,1,0))) );
    table->getMeshAt(0)->m_color = glm::vec4(0,0.310,0.235,1)*1.2f;
    table->getMeshAt(1)->m_color = glm::vec4(0.037,0.004,0.0,1)*3.f;
    table->getMeshAt(0)->m_shininess = 2.f;
    table->getMeshAt(1)->m_shininess = 5.f;
    table->getMeshAt(0)->m_specularStrength = 0.1f;
    table->getMeshAt(1)->m_specularStrength = 1.0f;
    table->simulateDynamics = false;
    scene->bodies.push_back( table );

    // Place white ball at near side
    auto whiteball = std::make_shared<Body>( std::vector<const char*>{SphereObjFile.c_str()}, true );
    whiteball->setScale(glm::vec3(BallRadius));
    whiteball->setPosition(glm::vec3(0,-2.2,0));
    whiteball->m_radius = BallRadius;
    whiteball->getMeshAt(0)->m_color = glm::vec4(0.6f,0.6f,0.6f,1.f);
    whiteball->getMeshAt(0)->m_shininess = 8.f;
    whiteball->getMeshAt(0)->m_specularStrength = 1.0f;
    scene->bodies.push_back( whiteball );

    // Place triangle of colored balls at far side
    for (int i=0; i<5; ++i) {
        for (int j=0; j<i+1; ++j) {
            auto ball = std::make_shared<Body>( std::vector<const char*>{SphereObjFile.c_str()}, true );
            ball->setScale(glm::vec3(BallRadius));
            if (i==2 && j==1)
                ball->getMeshAt(0)->m_color = glm::vec4(0.06,0.05,0.06,1);
            else
                ball->getMeshAt(0)->m_color = glm::normalize(glm::vec4((float)(std::rand())/(float)(RAND_MAX)
                                                                      ,(float)(std::rand())/(float)(RAND_MAX)
                                                                      ,(float)(std::rand())/(float)(RAND_MAX),1));
            static constexpr float BallExtraSpace = 0.001f;
            static constexpr float r = BallRadius + BallExtraSpace/2.f;
            static constexpr float d = BallDiameter + BallExtraSpace;
            ball->setPosition( glm::vec3(-i*r+j*d, 2-0.5+i*(d-(sqrtf(r*r+d*d)-d)), 0) );
            ball->m_radius = BallRadius;
            ball->getMeshAt(0)->m_shininess = 8.f;
            ball->getMeshAt(0)->m_specularStrength = 1.0f;
            scene->bodies.push_back( ball );
        }
    }

    // Start simulation
    Camera camera;
    Physics physics;
    physics.startSimulation( scene );

    // Reference old mouse position to determine change in position => dragging
    glm::vec2 lastMousePos(0,0);
    bool leftMouseUpBefore = true;
    bool rightMouseUpBefore = true;
    bool spaceUpBefore = true;
    float mx = -90;
    float my = 45;

    // Shooting auxiliary variables
    bool shooting = false;
    float shootingPower = 5.f;

    // We use a simple phong shader for all geometry
    Shader shader(CMAKE_SOURCE_DIR "/shader/phong.vs", CMAKE_SOURCE_DIR "/shader/phong.fs");

    // Main loop which is executed every frame until the window is closed
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();
        }

        // Setup the GL viewport
        glViewport(0, 0, display_w, display_h);
        glClearColor(0.05f, 0.05f, 0.05f, 1.f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Calculate aspect ratio (might be used for perspective transformation):
        float aspectRatio = float(display_w) / display_h;

        // Create view matrix from Camera member function
        if (physics.m_ballsMoving)
            camera.m_cameraMode = CameraMode::BirdsEye;
        else if (shooting)
            camera.m_cameraMode = CameraMode::Shooting;
        else
            camera.m_cameraMode = CameraMode::Trackball;
        glm::mat4 view = camera.getViewMatrix( whiteball, mx, my );

        // Interactive controls
        bool leftMouseDown = ImGui::IsMouseDown(ImGuiMouseButton_Left);
        bool rightMouseDown = ImGui::IsMouseDown(ImGuiMouseButton_Right);
        bool spaceDown = ImGui::IsKeyDown(ImGuiKey_Space);
        if (!physics.m_ballsMoving && !ImGui::GetIO().WantCaptureMouse) {
            auto m = ImGui::GetMousePos();
            glm::vec2 mousePos(m.x, m.y);
            auto deltaMouse = mousePos - lastMousePos;
            lastMousePos = mousePos;

            // Moving the camera view around
            if (leftMouseDown) {
                static constexpr float MouseSensitivity = 0.1f;
                static constexpr float CameraEpsilon = 0.001f;
                mx = mx - deltaMouse.x*MouseSensitivity;
                my = std::clamp<float>(my - deltaMouse.y*MouseSensitivity, CameraEpsilon, 90.f-CameraEpsilon);
            }

            // Switching between view options
            if (!rightMouseDown && !rightMouseUpBefore)
                shooting = !shooting;

            // Shooting the white ball
            if (!spaceDown && !spaceUpBefore) {
                // Extracting the view direction from the view matrix's z column
                glm::vec3 shootingDir( view[0][2], view[1][2], 0 );
                whiteball->setVelocity(-glm::normalize(shootingDir)*shootingPower);
            }

            // Button state changes
            if (leftMouseDown) leftMouseUpBefore = false;
            else leftMouseUpBefore = true;
            if (rightMouseDown) rightMouseUpBefore = false;
            else rightMouseUpBefore = true;
            if (spaceDown) spaceUpBefore = false;
            else spaceUpBefore = true;
        }

        // Mouse scrolling
        static constexpr float ScrollCameraSensitivity = 0.5f;
        static constexpr float MinCameraDistance = 1.f;
        static constexpr float MaxCameraDistance = 20.f;
        static constexpr float ScrollShootingSensitivity = 0.1f;
        static constexpr float MaxShootingPower = 20.f;
        // Update camera distance when scrolling the mouse:
        if (!physics.m_ballsMoving && !shooting && ImGui::GetIO().MouseWheel != 0)
            camera.m_radius = std::clamp<float>(camera.m_radius - ImGui::GetIO().MouseWheel*ScrollCameraSensitivity, MinCameraDistance, MaxCameraDistance);
        // While in shooting view we adjust shooting power
        if (!physics.m_ballsMoving && shooting && ImGui::GetIO().MouseWheel != 0)
            shootingPower = std::clamp<float>(shootingPower + ImGui::GetIO().MouseWheel*ScrollShootingSensitivity, ScrollShootingSensitivity, MaxShootingPower);

        // Shooting control window
        if (!physics.m_ballsMoving && shooting) {
            ImGui::SetNextWindowPos(ImVec2(0,0));
            ImGui::SetNextWindowSize(ImVec2(200,60+200));
            if (ImGui::Begin("Shooting control", nullptr, ImGuiWindowFlags_NoCollapse|ImGuiWindowFlags_NoResize)) {
                ImGui::VSliderFloat("##something", ImVec2(40,200), &shootingPower, ScrollShootingSensitivity, MaxShootingPower, "%.1f", 0);
                ImGui::End();
            }
        }

        // Debug window
        static constexpr float DebugWindowWidth = 500;
        static constexpr float DebugWindowHeight = 100;
        float adjustedHeight = display_h-DebugWindowHeight;
#if defined(__APPLE__)
        adjustedHeight = display_h/2.f-DebugWindowHeight;
#endif
        ImGui::SetNextWindowPos(ImVec2(0,adjustedHeight));
        ImGui::SetNextWindowSize(ImVec2(DebugWindowWidth,DebugWindowHeight));
        if (ImGui::Begin("Debug", nullptr, ImGuiWindowFlags_NoCollapse|ImGuiWindowFlags_NoResize)) {
            ImGui::Text("Trackball: %i,%i\n", (int)std::round(mx), (int)std::round(my));
            ImGui::Text("Physics frequency: %i Hz [%f]", (int)std::round(1.f/physics.dt), physics.dt);
            ImGui::End();
        }

        // Update shader state
        static constexpr float Fov = 30; // Low FOV for less perspective distortion
        glm::mat4 projection = glm::perspective( glm::radians(Fov), aspectRatio, 0.01f, 1000.f);
        shader.setUniform("projection", projection);
        shader.setUniform("view", view);
        // Extracting the view direction
        glm::vec3 center_eye( view[0][2], view[1][2], view[2][2] );
        // Extracting the view distance
        float center_eye_dot_eye = view[3][2];
        // Technically this is not 100% correct, but the error is rather low
        // FIXME: The correct value can be retrieved from the Camera class
        glm::vec3 view_pos = -center_eye*center_eye_dot_eye;
        shader.setUniform("viewPos", view_pos);
        shader.setUniform("lightColor", glm::vec3(0.925f,0.9f,1.0f));
        shader.setUniform("lightPos", glm::vec3(0,0,5.f));

        // Render all meshes
        for (auto body : scene->bodies)
            body->render( shader );


        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    physics.abortSimulation();
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}